@@ -1,5 +1,5 @@
 /*
-    Copyright (C) 2001-2006  Ben Kibbey <bjk@luxsci.net>
+    Copyright (C) 2001-2013 Ben Kibbey <bjk@luxsci.net>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,12 +13,26 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
+
+#ifdef HAVE_LIMITS_H
 #include <limits.h>
+#ifndef LINE_MAX
+#ifdef _POSIX2_LINE_MAX
+#define LINE_MAX _POSIX2_LINE_MAX
+#else
+#define LINE_MAX 2048
+#endif
+#endif
+#endif
 
 #define err(eval, fmt, ...)	print_string(eval, 1, fmt, ## __VA_ARGS__)
 #define errx(eval, fmt, ...)	print_string(eval, 0, fmt, ## __VA_ARGS__)
@@ -45,6 +59,4 @@ void print_string(int eval, int prog, const char *fmt, ...)
 
     if (eval > -1)
 	exit(eval);
-
-    return;
 }
