@@ -1,5 +1,5 @@
 /*
-    Copyright (C) 2001-2006  Ben Kibbey <bjk@luxsci.net>
+    Copyright (C) 2001-2013 Ben Kibbey <bjk@luxsci.net>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,8 +13,12 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -24,13 +28,17 @@
 #include <time.h>
 #include <pwd.h>
 #include <ctype.h>
-#include <utmp.h>
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#ifndef LINE_MAX
+#ifdef _POSIX2_LINE_MAX
+#define LINE_MAX _POSIX2_LINE_MAX
+#else
+#define LINE_MAX 2048
+#endif
+#endif
 #endif
-
-#include "login.h"
 
 #ifndef HAVE_STRSEP
 #include "../strsep.c"
@@ -40,23 +48,32 @@
 #include "../err.c"
 #endif
 
-void ui_module_init(int *chainable)
-{
-#ifdef DEBUG
-    fprintf(stderr, "%s: ui_module_init()\n", __FILE__);
+#include "login.h"
+
+#define LOGIN_OPTION_ORDER	"pdimyhtl"
+#define LOGIN_OPTION_STRING	"Lpdimyhtl:"
+
+static char options[9];		/* NULL terminated. */
+static char *last_options;
+static char **strings;
+static time_t now;
+static int login_count;
+#if (!defined(__FreeBSD_version) || __FreeBSD_version < 900000)
+static int lastlogfd;
 #endif
 
+void add_string(char ***, const char *);
+char *stamp(time_t, const char *);
+char *safe_strncat(char *, const char *, size_t);
+
+void ui_module_init(int *chainable)
+{
     *chainable = 0;
     time(&now);
-    return;
 }
 
 void ui_module_exit()
 {
-#ifdef DEBUG
-    fprintf(stderr, "%s: ui_module_exit()\n", __FILE__);
-#endif
-
 #ifdef HAVE_PROCFS
     if (procdir)
 	closedir(procdir);
@@ -67,21 +84,32 @@ void ui_module_exit()
 	kvm_close(kd);
 #endif
 
+#if (!defined(__FreeBSD_version) || __FreeBSD_version < 900000)
     if (lastlogfd)
 	close(lastlogfd);
-
-    return;
+#endif
 }
 
-#ifndef UTMPX_FORMAT
+static void free_logins(UTMP **u)
+{
+    if (login_count) {
+	UTMP **up;
+
+	for (up = u; *up; up++)
+	    free(*up);
+
+	free(u);
+    }
+}
+
+#ifndef HAVE_UTMPX_H
 /* This is for *BSD (login process id). */
 #ifdef BSD_KVM
-char *ui_module_pid(uid_t uid, int multi)
+static char *get_pid(uid_t uid, int multi)
 {
     static int firstrun;
     static char line[LINE_MAX];
     int cnt, i;
-    pid_t pid = 0;
     char errbuf[LINE_MAX];
     struct kinfo_proc *kp;
 
@@ -93,42 +121,57 @@ char *ui_module_pid(uid_t uid, int multi)
     if (!kd) {
 	firstrun = 1;
 
-#ifdef __NetBSD__
-	if ((kd = kvm_openfiles(NULL, NULL, NULL,
+#if defined(__NetBSD__) || defined (__OpenBSD__)
+	if ((kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, errbuf)) == NULL) {
 #else
 	if ((kd = kvm_openfiles(_PATH_DEVNULL, _PATH_DEVNULL, _PATH_DEVNULL,
-#endif
 				O_RDONLY, errbuf)) == NULL) {
+#endif
 	    warnx("%s", errbuf);
 	    return "!";
 	}
     }
 
+#ifdef __OpenBSD__
+    if ((kp = kvm_getprocs(kd, KERN_PROC_UID, uid, sizeof(struct kinfo_proc),
+	 &cnt)) == NULL) {
+	warnx("kvm_getprocs(): %s", kvm_geterr(kd));
+	return "!";
+    }
+#else
     if ((kp = kvm_getprocs(kd, KERN_PROC_UID, uid, &cnt)) == NULL) {
 	warnx("kvm_getprocs(): %s", kvm_geterr(kd));
 	return "!";
     }
+#endif
 
     for (i = 0; i < cnt; i++) {
 	char buf[32];
+	pid_t pid = 0;
 
+#ifdef __OpenBSD__
+	if ((kp[i].p_eflag & EPROC_SLEADER) && kp[i].p_tdev != -1)
+	    pid = kp[i].p_pid;
+#else
 #if __FreeBSD_version < 500000
-	if (kp[i].kp_eproc.e_flag & EPROC_SLEADER && kp[i].kp_eproc.e_tdev !=
-	    -1) {
+	if (kp[i].kp_eproc.e_flag & EPROC_SLEADER && kp[i].kp_eproc.e_tdev
+	    != -1) {
 	    pid = kp[i].kp_eproc.e_ppid;
 	    /*
 	     * pid = kp[i].kp_proc.p_pid;
 	     */
-
-	    if (pid == 1)
-		continue;
+	}
 #else
 	if (kp[i].ki_kiflag & KI_SLEADER && kp[i].ki_tdev != -1) {
 	    pid = kp[i].ki_pid;
-#endif
-	    snprintf(buf, sizeof(buf), "%i%c", pid, multi);
-	    strncat(line, buf, sizeof(line));
 	}
+#endif
+#endif
+	if (!pid || pid == 1)
+	    continue;
+
+        snprintf(buf, sizeof(buf), "%i%c", pid, multi);
+	safe_strncat(line, buf, sizeof(line));
     }
 
     if (line[0] == '\0')
@@ -147,21 +190,21 @@ char *ui_module_pid(uid_t uid, int multi)
 #include <dirent.h>
 #endif
 
-#ifdef __svr4__
+#ifdef __sun__
 #include <unistd.h>
 #include <procfs.h>
 #endif
 
-char *ui_module_pid(uid_t uid, int multi)
+static char *get_pid(uid_t uid, int multi)
 {
     static int firstrun;
     struct dirent *ent;
     struct stat st;
     static char line[LINE_MAX];
-    pid_t *pids = 0;
+    pid_t *pids = 0, *tpids;
     int pid_index = 0;
 
-#ifdef __svr4__
+#ifdef __sun__
     int fd;
     struct pstatus pstat;
 #else
@@ -184,14 +227,14 @@ char *ui_module_pid(uid_t uid, int multi)
 
     rewinddir(procdir);
 
-  again:
+again:
     while ((ent = readdir(procdir)) != NULL) {
 	pid_t pid = -1;
 	char filename[FILENAME_MAX];
 	char buf[LINE_MAX];
 	int i;
 
-#ifndef __svr4__
+#ifndef __sun__
 	char *t;
 #endif
 
@@ -208,10 +251,10 @@ char *ui_module_pid(uid_t uid, int multi)
 	    continue;
 
 	/*
-	 * The current user owns this file (process id). 
+	 * The current user owns this file (process id).
 	 */
 	if (st.st_uid == uid) {
-#ifdef __svr4__
+#ifdef __sun__
 	    if ((fd = open(filename, O_RDONLY)) == -1)
 		continue;
 
@@ -248,7 +291,7 @@ char *ui_module_pid(uid_t uid, int multi)
 #endif
 
 	    /*
-	     * Skip duplicate pids. 
+	     * Skip duplicate pids.
 	     */
 	    for (i = 0; i < pid_index; i++) {
 		if (pids[i] == pid)
@@ -256,14 +299,15 @@ char *ui_module_pid(uid_t uid, int multi)
 	    }
 
 	    snprintf(buf, sizeof(buf), "%li%c", (unsigned long) pid, multi);
-	    strncat(line, buf, sizeof(line));
+	    safe_strncat(line, buf, sizeof(line));
 
-	    if ((pids =
+	    if ((tpids =
 		 realloc(pids, (pid_index + 2) * sizeof(pid_t *))) == NULL) {
 		warn("realloc()");
 		continue;
 	    }
 
+	    pids = tpids;
 	    pids[pid_index++] = pid;
 	}
     }
@@ -279,7 +323,7 @@ char *ui_module_pid(uid_t uid, int multi)
 }
 #else
 /* Unsupported OS. */
-char *ui_module_pid(uid_t uid, int multi)
+static char *get_pid(uid_t uid, int multi)
 {
     return "!";
 }
@@ -334,30 +378,56 @@ static void last_strings(char *str)
 		break;
 	}
     }
-
-    return;
 }
 
 /* Get the lastlog structure from the lastlog file. */
-static char *lastlogin(uid_t uid, char *tf)
+#ifdef HAVE_GETLASTLOGX
+static char *lastlogin(const struct passwd *pw, char *tf)
+{
+    struct lastlogx *last =  getlastlogx(_PATH_LASTLOGX, pw->pw_uid, NULL);
+    static char buf[LINE_MAX];
+
+    if (!last)
+	return NULL;
+
+    snprintf(buf, sizeof(buf), "%s,%s,%s",
+	!last->ll_line[0] ? "!" : last->ll_line,
+	(!last->ll_host[0] || !isalnum(last->ll_host[0])) ?  "-" : last->ll_host,
+	stamp(last->ll_tv.tv_sec, tf));
+    return buf;
+}
+#else
+#if __FreeBSD_version >= 900000
+static char *lastlogin(const struct passwd *pw, char *tf)
+{
+    struct utmpx *last;
+    static char buf[LINE_MAX];
+
+    if (setutxdb(UTXDB_LASTLOGIN, NULL) == -1) {
+	warn("lastlog");
+	return NULL;
+    }
+
+    last = getutxuser(pw->pw_name);
+
+    if (!last)
+	return NULL;
+
+    snprintf(buf, sizeof(buf), "%s,%s,%s",
+	!last->ut_line[0] ? "!" : last->ut_line,
+	(!last->ut_host[0] || !isalnum(last->ut_host[0])) ?  "-" : last->ut_host,
+	stamp(last->ut_tv.tv_sec, tf));
+
+    return buf;
+}
+#else
+static char *lastlogin(const struct passwd *pw, char *tf)
 {
     int count;
     long offset;
     static char buf[LINE_MAX];
-
-#ifdef __NetBSD__
-#ifdef UTMPX_FORMAT
-    char tmp[64], htmp[UTX_HOSTSIZE + 1];
-#else
-    char tmp[64], htmp[UT_HOSTSIZE + 1];
-#endif
-#else
-    char tmp[64], htmp[UT_HOSTSIZE + 1];
-#endif
     struct lastlog last;
 
-    buf[0] = tmp[0] = htmp[0] = '\0';
-
     if (lastlogfd < 0)
 	return NULL;
 
@@ -368,7 +438,7 @@ static char *lastlogin(uid_t uid, char *tf)
 	}
     }
 
-    offset = (long) uid *sizeof(struct lastlog);
+    offset = (long) pw->pw_uid * sizeof(struct lastlog);
 
     if (lseek(lastlogfd, offset, SEEK_SET) == -1) {
 	warn("%s", _PATH_LASTLOG);
@@ -383,46 +453,28 @@ static char *lastlogin(uid_t uid, char *tf)
 	return NULL;
     }
 
-    if (last.ll_line[0] == '\0')
-	strncpy(buf, "!", sizeof(buf));
-    else
-	strncpy(buf, last.ll_line, sizeof(buf));
-
-    strncat(buf, ",", sizeof(buf));
-    strncpy(htmp, last.ll_host, sizeof(htmp));
-
 #ifdef __NetBSD__
-#ifdef UTMPX_FORMAT
-    htmp[UTX_HOSTSIZE] = '\0';
+#ifdef HAVE_UTMPX_H
+    last.ll_host[UTX_HOSTSIZE-1] = '\0';
+    last.ll_line[UTX_LINESIZE-1] = '\0';
 #else
-    htmp[UT_HOSTSIZE] = '\0';
+    last.ll_host[UT_HOSTSIZE-1] = '\0';
+    last.ll_line[UT_LINESIZE-1] = '\0';
 #endif
 #else
-    htmp[UT_HOSTSIZE] = '\0';
+    last.ll_host[UT_HOSTSIZE-1] = '\0';
+    last.ll_line[UT_LINESIZE-1] = '\0';
 #endif
 
-    if (htmp[0] && isalnum((unsigned char) htmp[0]))
-	strncat(buf, htmp, sizeof(buf));
-    else {
-	/*
-	 * If a users tty is tty1-n, it must be a console
-	 * * login. 
-	 */
-	if (last.ll_line[0] && isdigit((unsigned char) last.ll_line[3]))
-	    strncat(buf, "-", sizeof(buf));
-	else
-	    strncat(buf, "!", sizeof(buf));
-    }
-
-    strncat(buf, ",", sizeof(buf));
-
-    if (last.ll_time)
-	strncat(buf, stamp(last.ll_time, tf), sizeof(buf));
-    else
-	strncat(buf, "!", sizeof(buf));
-
+    snprintf(buf, sizeof(buf), "%s,%s,%s",
+	!last.ll_line[0] ? "!" : last.ll_line,
+	(!last.ll_host[0] || !isalnum(last.ll_host[0])) ?
+	    !isdigit(last.ll_line[3]) ? "!" : "-" : last.ll_host,
+	!last.ll_time ? "!" : stamp(last.ll_time, tf));
     return buf;
 }
+#endif
+#endif
 
 /* This will return an array of utmp structures if a user is logged in, NULL
  * otherwise. We'll try to keep the utmp file descriptor open if possible to
@@ -430,8 +482,7 @@ static char *lastlogin(uid_t uid, char *tf)
 static UTMP **get_utmp(const char *user)
 {
     UTMP **logins = NULL;
-
-#ifdef UTMPX_FORMAT
+#if defined HAVE_UTMPX_H && defined (HAVE_SETUTXENT)
     UTMP *u;
 #else
     UTMP u;
@@ -451,47 +502,61 @@ static UTMP **get_utmp(const char *user)
 
     login_count = 0;
 
-#ifdef UTMPX_FORMAT
-    setutent();
+#if defined HAVE_UTMPX_H && defined (HAVE_SETUTXENT)
+    setutxent();
 
     while ((u = getutxent()) != NULL) {
-	if (strcmp(u->ut_name, user) == 0) {
+	if (!strcmp(u->ut_user, user) && u->ut_type != DEAD_PROCESS) {
 #else
     lseek(fd, 0, SEEK_SET);
 
     while ((count = read(fd, &u, sizeof(UTMP))) == sizeof(UTMP)) {
 	if (strcmp(u.ut_name, user) == 0) {
 #endif
-	    if ((logins = realloc(logins,
+	    UTMP **tmp;
+
+	    if ((tmp = realloc(logins,
 				  (login_count + 2) * sizeof(UTMP *))) ==
 		NULL) {
 		warn("realloc()");
+		free_logins(logins);
 		return NULL;
 	    }
 
+	    logins = tmp;
+
 	    if ((logins[login_count] = malloc(sizeof(UTMP))) == NULL) {
 		warn("malloc()");
+		free_logins(logins);
 		return NULL;
 	    }
 
-#ifdef UTMPX_FORMAT
+#if defined HAVE_UTMPX_H && defined (HAVE_SETUTXENT)
 #ifdef __NetBSD__
-	    strncpy(logins[login_count]->ut_name, u->ut_name, UTX_NAMESIZE);
-	    strncpy(logins[login_count]->ut_line, u->ut_line, UTX_LINESIZE);
-	    strncpy(logins[login_count]->ut_host, u->ut_host, UTX_HOSTSIZE);
+	    memcpy(logins[login_count]->ut_name, u->ut_name, UTX_NAMESIZE);
+	    logins[login_count]->ut_name[UTX_NAMESIZE-1] = 0;
+	    memcpy(logins[login_count]->ut_line, u->ut_line, UTX_LINESIZE);
+	    logins[login_count]->ut_line[UTX_LINESIZE-1] = 0;
+	    memcpy(logins[login_count]->ut_host, u->ut_host, UTX_HOSTSIZE);
+	    logins[login_count]->ut_host[UTX_HOSTSIZE-1] = 0;
 	    logins[login_count]->ut_pid = u->ut_pid;
 #else
-	    strncpy(logins[login_count]->ut_name, u->ut_name, UT_NAMESIZE);
-	    strncpy(logins[login_count]->ut_line, u->ut_line, UT_LINESIZE);
-	    strncpy(logins[login_count]->ut_host, u->ut_host, UT_HOSTSIZE);
+	    memcpy(logins[login_count]->ut_user, u->ut_user, UT_NAMESIZE);
+	    logins[login_count]->ut_user[UT_NAMESIZE-1] = 0;
+	    memcpy(logins[login_count]->ut_line, u->ut_line, UT_LINESIZE);
+	    logins[login_count]->ut_line[UT_LINESIZE-1] = 0;
+	    memcpy(logins[login_count]->ut_host, u->ut_host, UT_HOSTSIZE);
+	    logins[login_count]->ut_host[UT_HOSTSIZE-1] = 0;
 	    logins[login_count]->ut_tv.tv_sec = u->ut_tv.tv_sec;
 	    logins[login_count]->ut_pid = u->ut_pid;
 #endif
 #else
-	    strncpy(logins[login_count]->ut_name, u.ut_name, UT_NAMESIZE);
-	    strncpy(logins[login_count]->ut_line, u.ut_line, UT_LINESIZE);
-	    strncpy(logins[login_count]->ut_host, u.ut_host, UT_HOSTSIZE);
-	    logins[login_count]->ut_host[UT_HOSTSIZE] = '\0';
+	    memcpy(logins[login_count]->ut_name, u.ut_name, UT_NAMESIZE);
+	    logins[login_count]->ut_name[UT_NAMESIZE-1] = 0;
+	    memcpy(logins[login_count]->ut_line, u.ut_line, UT_LINESIZE);
+	    logins[login_count]->ut_line[UT_LINESIZE-1] = 0;
+	    memcpy(logins[login_count]->ut_host, u.ut_host, UT_HOSTSIZE);
+	    logins[login_count]->ut_host[UT_HOSTSIZE-1] = 0;
 	    logins[login_count]->ut_time = u.ut_time;
 #endif
 	    logins[++login_count] = NULL;
@@ -502,7 +567,7 @@ static UTMP **get_utmp(const char *user)
 }
 
 /* The 'mesg' status of the logged in user. */
-static char *msgstat(UTMP ** u, int multi)
+static char *msgstat(UTMP **u, int multi)
 {
     static char line[LINE_MAX];
     int i;
@@ -517,13 +582,13 @@ static char *msgstat(UTMP ** u, int multi)
 	snprintf(filename, sizeof(filename), "%s%s", _PATH_DEV, u[i]->ut_line);
 
 	if (stat(filename, &st) == -1)
-	    strncat(line, "!", sizeof(line));
+	    safe_strncat(line, "!", sizeof(line));
 	else
-	    strncat(line,
+	    safe_strncat(line,
 		    (st.st_mode & S_IWGRP || st.st_mode & S_IWOTH) ? "1" : "0",
 		    sizeof(line));
 
-	strncat(line, m, sizeof(line));
+	safe_strncat(line, m, sizeof(line));
     }
 
     if (line[0] == '\0')
@@ -534,7 +599,7 @@ static char *msgstat(UTMP ** u, int multi)
 }
 
 /* Returns the users idle time in seconds. */
-static char *idle(UTMP ** u, int multi)
+static char *idle(UTMP **u, int multi)
 {
     static char line[LINE_MAX];
     time_t t;
@@ -550,24 +615,24 @@ static char *idle(UTMP ** u, int multi)
 	snprintf(buf, sizeof(buf), "%s%s", _PATH_DEV, u[i]->ut_line);
 
 	if (stat(buf, &st) == -1) {
-	    strncat(line, "!", sizeof(line));
-	    strncat(line, m, sizeof(line));
+	    safe_strncat(line, "!", sizeof(line));
+	    safe_strncat(line, m, sizeof(line));
 	    continue;
 	}
 
-#ifdef UTMPX_FORMAT
+#ifdef HAVE_UTMPX_H
 	if (u[i]->ut_tv.tv_sec > st.st_atime) {
 #else
 	if (u[i]->ut_time > st.st_atime) {
 #endif
-	    strncat(line, "-", sizeof(line));
-	    strncat(line, m, sizeof(line));
+	    safe_strncat(line, "-", sizeof(line));
+	    safe_strncat(line, m, sizeof(line));
 	    continue;
 	}
 
 	t = st.st_atime;
 
-#ifdef UTMPX_FORMAT
+#ifdef HAVE_UTMPX_H
 	if (t < u[i]->ut_tv.tv_sec)
 	    t = u[i]->ut_tv.tv_sec;
 #else
@@ -576,8 +641,8 @@ static char *idle(UTMP ** u, int multi)
 #endif
 
 	snprintf(buf, sizeof(buf), "%lu", (now - t <= 0) ? 0 : now - t);
-	strncat(line, buf, sizeof(line));
-	strncat(line, m, sizeof(line));
+	safe_strncat(line, buf, sizeof(line));
+	safe_strncat(line, m, sizeof(line));
     }
 
     if (line[0] == '\0')
@@ -591,10 +656,6 @@ static char *idle(UTMP ** u, int multi)
  */
 void ui_module_help()
 {
-#ifdef DEBUG
-    fprintf(stderr, "%s: ui_module_help()\n", __FILE__);
-#endif
-
     printf("  Login information [-L (-%s)]:\n", LOGIN_OPTION_ORDER);
     printf("\t-y  tty\t\t\t\t");
     printf("-m  message status\n");
@@ -613,7 +674,7 @@ int ui_module_exec(char ***s, const struct passwd *pw, const int multi,
 	       const int verbose, char *tf)
 {
     char *p = options;
-    UTMP **u = NULL, **up;
+    UTMP **u = NULL;
     char buf[255];
 
     login_count = 0;
@@ -624,44 +685,23 @@ int ui_module_exec(char ***s, const struct passwd *pw, const int multi,
 	char line[LINE_MAX] = { '\0' };
 	int i;
 	char m[2] = { multi, '\0' };
-#ifdef __NetBSD__
-#ifdef UTMPX_FORMAT
-	char htmp[UTX_HOSTSIZE + 1];
-#else
-	char htmp[UT_HOSTSIZE + 1];
-#endif
-#else
-	char htmp[UT_HOSTSIZE + 1];
-#endif
 
 	switch (*p) {
 	    case 'i':
 		add_string(&strings, (u) ? idle(u, multi) : "!");
 		break;
 	    case 'l':
-		last_strings(lastlogin(pw->pw_uid, tf));
+		last_strings(lastlogin(pw, tf));
 		break;
 	    case 'h':
 		for (i = 0; i < login_count; i++) {
 		    if (u[i]->ut_host[0]
-			&& isalnum((unsigned char) u[i]->ut_host[0])) {
-			strcpy(htmp, u[i]->ut_host);
-			htmp[sizeof(htmp) - 1] = '\0';
-			strncat(line, htmp, sizeof(line));
-		    }
-		    else {
-			/*
-			 * If a users tty is tty1-n, it must be a console
-			 * * login. 
-			 */
-			if (u[i]->ut_line[0]
-			    && isdigit((unsigned char) u[i]->ut_line[3]))
-			    strncat(line, "-", sizeof(line));
-			else
-			    strncat(line, "!", sizeof(line));
-		    }
+			    && isalnum((unsigned char) u[i]->ut_host[0]))
+			safe_strncat(line, u[i]->ut_host, sizeof(line));
+		    else
+			safe_strncat(line, "-", sizeof(line));
 
-		    strncat(line, m, sizeof(line));
+		    safe_strncat(line, m, sizeof(line));
 		}
 
 		if (line[0] == '\0')
@@ -674,11 +714,11 @@ int ui_module_exec(char ***s, const struct passwd *pw, const int multi,
 	    case 'y':
 		for (i = 0; i < login_count; i++) {
 		    if (u[i]->ut_line[0])
-			strncat(line, u[i]->ut_line, sizeof(line));
+			safe_strncat(line, u[i]->ut_line, sizeof(line));
 		    else
-			strncat(line, "!", sizeof(line));
+			safe_strncat(line, "!", sizeof(line));
 
-		    strncat(line, m, sizeof(line));
+		    safe_strncat(line, m, sizeof(line));
 		}
 
 		if (line[0] == '\0')
@@ -693,12 +733,12 @@ int ui_module_exec(char ***s, const struct passwd *pw, const int multi,
 		break;
 	    case 't':
 		for (i = 0; i < login_count; i++) {
-#ifdef UTMPX_FORMAT
-		    strncat(line, stamp(u[i]->ut_tv.tv_sec, tf), sizeof(line));
+#ifdef HAVE_UTMPX_H
+		    safe_strncat(line, stamp(u[i]->ut_tv.tv_sec, tf), sizeof(line));
 #else
-		    strncat(line, stamp(u[i]->ut_time, tf), sizeof(line));
+		    safe_strncat(line, stamp(u[i]->ut_time, tf), sizeof(line));
 #endif
-		    strncat(line, m, sizeof(line));
+		    safe_strncat(line, m, sizeof(line));
 		}
 
 		if (line[0] == '\0')
@@ -710,7 +750,7 @@ int ui_module_exec(char ***s, const struct passwd *pw, const int multi,
 		break;
 	    case 'd':
 		for (i = 0; i < login_count; i++) {
-#ifdef UTMPX_FORMAT
+#ifdef HAVE_UTMPX_H
 		    if ((now - u[i]->ut_tv.tv_sec) > 60) {
 			snprintf(buf, sizeof(buf), "%lu",
 				 ((now - u[i]->ut_tv.tv_sec) / 60));
@@ -719,12 +759,12 @@ int ui_module_exec(char ***s, const struct passwd *pw, const int multi,
 			snprintf(buf, sizeof(buf), "%lu",
 				 ((now - u[i]->ut_time) / 60));
 #endif
-			strncat(line, buf, sizeof(line));
+			safe_strncat(line, buf, sizeof(line));
 		    }
 		    else
-			strncat(line, "-", sizeof(line));
+			safe_strncat(line, "-", sizeof(line));
 
-		    strncat(line, m, sizeof(line));
+		    safe_strncat(line, m, sizeof(line));
 		}
 
 		if (line[0] == '\0')
@@ -735,16 +775,16 @@ int ui_module_exec(char ***s, const struct passwd *pw, const int multi,
 		add_string(&strings, line);
 		break;
 	    case 'p':
-#ifdef UTMPX_FORMAT
+#ifdef HAVE_UTMPX_H
 		for (i = 0; i < login_count; i++) {
 		    if (u[i]->ut_pid) {
 			snprintf(buf, sizeof(buf), "%li", (long) u[i]->ut_pid);
-			strncat(line, buf, sizeof(line));
+			safe_strncat(line, buf, sizeof(line));
 		    }
 		    else
-			strncat(line, "!", sizeof(line));
+			safe_strncat(line, "!", sizeof(line));
 
-		    strncat(line, m, sizeof(line));
+		    safe_strncat(line, m, sizeof(line));
 		}
 
 		if (line[0] == '\0')
@@ -754,7 +794,7 @@ int ui_module_exec(char ***s, const struct passwd *pw, const int multi,
 
 		add_string(&strings, line);
 #else
-		add_string(&strings, (u) ? ui_module_pid(pw->pw_uid, multi) : "!");
+		add_string(&strings, (u) ? get_pid(pw->pw_uid, multi) : "!");
 #endif
 		break;
 	    default:
@@ -764,13 +804,7 @@ int ui_module_exec(char ***s, const struct passwd *pw, const int multi,
 	p++;
     }
 
-    if (login_count) {
-	for (up = u; *up; up++)
-	    free(*up);
-
-	free(u);
-    }
-
+    free_logins(u);
     *s = strings;
     return EXIT_SUCCESS;
 }
@@ -807,16 +841,18 @@ int ui_module_options(int argc, char **argv)
     int opt;
     char *p = options;
 
-#ifdef DEBUG
-    fprintf(stderr, "%s: ui_module_options()\n", __FILE__);
-#endif
-
     while ((opt = getopt(argc, argv, LOGIN_OPTION_STRING)) != -1) {
 	switch (opt) {
 	    case 'l':
+		if (parse_last_options(optarg))
+		    return 1;
+
 		last_options = optarg;
 		break;
 	    case 'L':
+		strncpy(options, LOGIN_OPTION_ORDER, sizeof(options));
+		last_options = "a";
+		return 0;
 	    case 'p':
 	    case 'd':
 	    case 'i':
@@ -831,17 +867,6 @@ int ui_module_options(int argc, char **argv)
 		return 1;
 	}
 
-	if (opt == 'l') {
-	    if (parse_last_options(last_options))
-		return 1;
-	}
-
-	if (opt == 'L') {
-	    strncpy(options, LOGIN_OPTION_ORDER, sizeof(options));
-	    last_options = "a";
-	    break;
-	}
-
 	*p++ = opt;
 	*p = '\0';
     }
