@@ -1,5 +1,5 @@
 /*
-    Copyright (C) 2001-2006  Ben Kibbey <bjk@luxsci.net>
+    Copyright (C) 2001-2013 Ben Kibbey <bjk@luxsci.net>
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -13,8 +13,12 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -25,8 +29,15 @@
 #include <pwd.h>
 #include <time.h>
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#ifndef LINE_MAX
+#ifdef _POSIX2_LINE_MAX
+#define LINE_MAX _POSIX2_LINE_MAX
+#else
+#define LINE_MAX 2048
+#endif
+#endif
 #endif
 
 #ifdef HAVE_GETOPT_H
@@ -37,6 +48,7 @@
 #include "err.c"
 #endif
 
+#include "safe_strncat.c"
 #include "ui.h"
 
 static void *Realloc(void *p, size_t size)
@@ -58,11 +70,10 @@ char *stamp(time_t epoch, const char *format)
 
     t = localtime(&epoch);
     strftime(buf, sizeof(buf), format, t);
-
     return buf;
 }
 
-/* 
+/*
  * This may be used in modules to add a string to the buffer (ui_module_exec()).
  */
 void add_string(char ***buf, const char *str)
@@ -80,7 +91,6 @@ void add_string(char ***buf, const char *str)
     s[i++] = strdup(str);
     s[i] = NULL;
     *buf = s;
-    return;
 }
 
 /* This is for the field separators (-F and -m). */
@@ -130,7 +140,6 @@ static void usage_header()
     printf("Usage: %s [-vhVL] [-c <filename>] [-t fmt] [-m c] [-F c] [-d]\n"
 	   "\t[[-xX] -O <module1> [options] [-- [-xX] -O <module2> [...]]]\n"
 	   "\t[- | username | -f filename] [...]\n\n", __progname);
-    return;
 }
 
 /* Help text. Module help text is displayed before this. */
@@ -153,7 +162,6 @@ static void usage()
     printf("  -V\tVersion information.\n\n");
     printf("Output key: %s=unknown/error, %s=none, %s=yes/on, "
 	   "%s=no/off\n", UNKNOWN, NONE, ON, OFF);
-    return;
 }
 
 /*
@@ -161,13 +169,14 @@ static void usage()
  * current item number being added stored in an integer. The module is also
  * initialized here with ui_module_init().
  */
-static int open_module(char *filename, int *idx)
+static int open_module(char *filename, struct module_s **ptr)
 {
     void *m;
     module_init *init;
-    char *p, s[FILENAME_MAX];
-    int i;
+    char *p, s[PATH_MAX];
     int chainable = 0;
+    struct module_s *mod, **mods, *last;
+    int dup = 0;
 
     strncpy(s, filename, sizeof(s));
 
@@ -178,13 +187,10 @@ static int open_module(char *filename, int *idx)
 	p = s;
     }
 
-    for (i = 0; i < module_index; i++) {
-	if (strcmp(p, modules[i].name) == 0) {
-	    if (TEST_FLAG(modules[i].flags, MODULE_DUP))
-		break;
-
-	    SET_FLAG(modules[i].flags, MODULE_DUP);
-	    warnx("%s: a module by this name is already loaded", p);
+    for (mods = modules; mods && *mods; mods++) {
+	if (strcmp(p, (*mods)->name) == 0) {
+	    dup = 1;
+	    break;
 	}
     }
 
@@ -195,40 +201,50 @@ static int open_module(char *filename, int *idx)
 	return 1;
     }
 
-    modules = Realloc(modules, (module_index + 2) * sizeof(struct module));
-    modules[module_index].m = m;
+    last = modules ? modules[module_total-1] : NULL;
+    modules = Realloc(modules, (module_total+2) * sizeof(struct module_s *));
+    mod = calloc(1, sizeof(struct module_s));
+    modules[module_total++] = mod;
+    modules[module_total] = NULL;
+    mod->m = m;
+    strncpy(mod->name, p, sizeof(mod->name));
 
-    strncpy(modules[module_index].name, p, sizeof(modules[module_index].name));
-    *idx = module_index++;
-
-    if ((init = dlsym(modules[*idx].m, "ui_module_init")) == NULL)
+    if ((init = dlsym(mod->m, "ui_module_init")) == NULL)
 	warnx("%s", dlerror());
     else
 	(*init) (&chainable);
 
     if (chainable)
-	SET_FLAG(modules[*idx].flags, MODULE_CHAINABLE);
+	SET_FLAG(mod->flags, MODULE_CHAINABLE);
 
-    if (*idx - 1 >= 0 && TEST_FLAG(modules[*idx - 1].flags, MODULE_CHAINED) &&
-	    !TEST_FLAG(modules[*idx].flags, MODULE_CHAINABLE)) {
-	warnx("%s: this module is not chainable", modules[*idx].name);
+    if (last && TEST_FLAG(last->flags, MODULE_CHAINED) &&
+	    !TEST_FLAG(mod->flags, MODULE_CHAINABLE)) {
+	warnx("%s: this module is not chainable", mod->name);
+	module_total--;
+	modules[module_total] = NULL;
+	free(mod);
 	return 1;
     }
 
     /* Module chaining. See junction() for more info. */
     if (chaining)
-	SET_FLAG(modules[*idx].flags, MODULE_CHAINED);
+	SET_FLAG(mod->flags, MODULE_CHAINED);
 
     if (chain_output)
-	SET_FLAG(modules[*idx].flags, MODULE_OUTPUT);
+	SET_FLAG(mod->flags, MODULE_OUTPUT);
 
     if (verbose)
-	SET_FLAG(modules[*idx].flags, MODULE_VERBOSE);
+	SET_FLAG(mod->flags, MODULE_VERBOSE);
+
+    if (dup) {
+	SET_FLAG(mod->flags, MODULE_DUP);
+	warnx("%s: a module by this name is already loaded", p);
+    }
 
     chaining = 0;
     chain_output = 1;
     verbose = (verbose < 2) ? 0 : 2;
-
+    *ptr = mod;
     return 0;
 }
 
@@ -236,52 +252,59 @@ static int open_module(char *filename, int *idx)
  * themselves via the ui_module_exit() function. */
 static void cleanup_modules()
 {
-    int i;
+    struct module_s **mod;
 
-    for (i = 0; i < module_index; i++) {
+    if (!modules)
+	return;
+
+    for (mod = modules; *mod; mod++) {
 	module_exit *e;
+	char **p;
 
-	if ((e = dlsym(modules[i].m, "ui_module_exit")) == NULL)
+	if ((e = dlsym((*mod)->m, "ui_module_exit")) == NULL)
 	    warnx("%s", dlerror());
 	else
-	    (*e) ();
+	    (*e)();
 
-	dlclose(modules[i].m);
+	dlclose((*mod)->m);
+
+	if ((*mod)->argv) {
+	    for (p = (*mod)->argv; *p; p++)
+		free(*p);
+
+	    free((*mod)->argv);
+	}
+
+	free(*mod);
     }
 
     free(modules);
-    return;
 }
 
 static void output(char **s, const int sep, int which)
 {
-    int i;
+    char **p = s;
 
-    if (s) {
-	for (i = 0; s[i]; i++) {
-	    printf("%s", s[i]);
+    if (p) {
+	for (; *p; p++) {
+	    printf("%s", *p);
 
-	    if (s[i + 1])
+	    if (*(p+1))
 		printf("%c", sep);
 	}
     }
 
-    if (which == OUTPUT_DONE)
-	printf("\n");
-    else if (which == OUTPUT_APPEND)
-	printf("%c", sep);
-
-    return;
+    printf("%c", which == OUTPUT_DONE ? '\n' : sep);
 }
 
 /* Pass the argument to each loaded module. */
 static int junction(const char *arg)
 {
-    int i;
     struct passwd *pw;
     struct stat st;
     int ret = EXIT_SUCCESS;
     char **s = NULL;
+    struct module_s **mod;
 
     if (usefile) {
 	if ((STAT(arg, &st)) == -1) {
@@ -323,22 +346,26 @@ static int junction(const char *arg)
 	}
     }
 
-    for (i = 0; i < module_index; i++) {
+    for (mod = modules; *mod; mod++) {
 	module_exec *m_exec;
-	char **p;
 
-	if ((m_exec = dlsym(modules[i].m, "ui_module_exec")) == NULL) {
+	if ((m_exec = dlsym((*mod)->m, "ui_module_exec")) == NULL) {
 	    warnx("%s", dlerror());
 	    continue;
 	}
 
-	ret |= (*m_exec) (&s, pw, multichar, TEST_FLAG(modules[i].flags, MODULE_VERBOSE), tf);
+	ret |= (*m_exec) (&s, pw, multichar,
+		TEST_FLAG((*mod)->flags, MODULE_VERBOSE), tf);
 
-	if (!TEST_FLAG(modules[i].flags, MODULE_CHAINED) || (TEST_FLAG(modules[i].flags, MODULE_CHAINED) && TEST_FLAG(modules[i].flags, MODULE_OUTPUT))) {
+	if (!TEST_FLAG((*mod)->flags, MODULE_CHAINED) ||
+		(TEST_FLAG((*mod)->flags, MODULE_CHAINED) &&
+		 TEST_FLAG((*mod)->flags, MODULE_OUTPUT))) {
 	    output(s, delimchar,
-		    ((i + 1) < module_index) ? OUTPUT_APPEND : OUTPUT_DONE);
+		    (*(mod+1)) ? OUTPUT_APPEND : OUTPUT_DONE);
+
+	    if (!TEST_FLAG((*mod)->flags, MODULE_CHAINED)) {
+		char **p;
 
-	    if (!TEST_FLAG(modules[i].flags, MODULE_CHAINED)) {
 		for (p = s; *p; p++) {
 		    free(*p);
 		}
@@ -354,29 +381,28 @@ static int junction(const char *arg)
 
 /* Copy options for each module into it's own argc and argv variables stopping
  * at -- (getopt(3)). */
-static int init_module_options(int the_argc, char **the_argv, struct module mod)
+static int init_module_options(int the_argc, char **the_argv,
+	struct module_s *mod)
 {
     char tmp[255];
     module_options *m;
     module_options_init *o;
     int old_optind = optind;
-    int argc = 0;
-    char **argv = NULL;
     int opt;
     int ret = EXIT_SUCCESS;
     char *optstring = NULL;
     char *defaults = NULL;
     int have_an_argument = 0;
 
-    if ((o = dlsym(mod.m, "ui_module_options_init")) == NULL) {
+    if ((o = dlsym(mod->m, "ui_module_options_init")) == NULL) {
 	warnx("%s", dlerror());
 	return EXIT_FAILURE;
     }
 
     if ((optstring = (*o) (&defaults))) {
-	argv = Realloc(argv, (argc + 2) * sizeof(char *));
-	argv[argc++] = strdup(__progname);
-	argv[argc] = NULL;
+	mod->argv = Realloc(mod->argv, (mod->argc + 2) * sizeof(char *));
+	mod->argv[mod->argc++] = strdup(__progname);
+	mod->argv[mod->argc] = NULL;
 
 	/* Probably a default module. */
 	if (the_argv == NULL)
@@ -385,17 +411,17 @@ static int init_module_options(int the_argc, char **the_argv, struct module mod)
 	while ((opt = getopt(the_argc, the_argv, optstring)) != -1) {
 	    switch (opt) {
 		case '?':
-		    warnx("%s: invalid option -- %c\n", mod.name,
+		    warnx("%s: invalid option -- %c\n", mod->name,
 			  optopt);
 		    return EXIT_FAILURE;
 		default:
 		    break;
 	    }
 
-	    argv = Realloc(argv, (argc + 2) * sizeof(char *));
+	    mod->argv = Realloc(mod->argv, (mod->argc + 2) * sizeof(char *));
 	    snprintf(tmp, sizeof(tmp), "-%c%s", opt, (optarg) ? optarg : "");
-	    argv[argc++] = strdup(tmp);
-	    argv[argc] = NULL;
+	    mod->argv[mod->argc++] = strdup(tmp);
+	    mod->argv[mod->argc] = NULL;
 	    have_an_argument = 1;
 	}
     }
@@ -408,21 +434,21 @@ blah:
      * options (ui_module_options_init()) if any. 
      */
     if (!have_an_argument && defaults) {
-	argv = Realloc(argv, (argc + 2) * sizeof(char *));
+	mod->argv = Realloc(mod->argv, (mod->argc + 2) * sizeof(char *));
 	snprintf(tmp, sizeof(tmp), "-%s", defaults);
-	argv[argc++] = strdup(tmp);
-	argv[argc] = NULL;
+	mod->argv[mod->argc++] = strdup(tmp);
+	mod->argv[mod->argc] = NULL;
     }
 
     old_optind = optind;
     opterr = optind = optopt = 1;
 
-    if ((m = dlsym(mod.m, "ui_module_options")) == NULL) {
+    if ((m = dlsym(mod->m, "ui_module_options")) == NULL) {
 	warnx("%s", dlerror());
 	return EXIT_FAILURE;
     }
 
-    ret |= (*m) (argc, argv);
+    ret |= (*m) (mod->argc, mod->argv);
     optind = old_optind;
 
 skip_option_stuff:
@@ -448,7 +474,7 @@ skip_option_stuff:
 static char **parseargv(char *str, const char *progname, int *me_argc)
 {
     char **pptr, *s;
-    char arg[255];
+    char arg[LINE_MAX];
     int idx = 0;
     int quote = 0;
     int lastchar = 0;
@@ -502,7 +528,7 @@ static char **parseargv(char *str, const char *progname, int *me_argc)
 static char *get_home_directory()
 {
     struct passwd *pw;
-    static char dir[FILENAME_MAX];
+    static char dir[PATH_MAX];
 
     errno = 0;
 
@@ -525,8 +551,8 @@ static int parse_rc_file(const char *filename)
 {
     char line[LINE_MAX], *p;
     FILE *fp;
-    int idx;
     int old_optind = optind;
+    struct module_s *mod;
     
     if ((fp = fopen(filename, "r")) == NULL) {
 	warn("%s", filename);
@@ -534,10 +560,11 @@ static int parse_rc_file(const char *filename)
     }
 
     while ((p = fgets(line, sizeof(line), fp)) != NULL) {
-	char name[FILENAME_MAX], options[LINE_MAX], tmp[LINE_MAX], *s;
+	char name[PATH_MAX], options[LINE_MAX], tmp[PATH_MAX], *s;
 	int my_argc;
-	char **my_argv;
+	char **my_argv, **ap;
 	int lastchar = '\0';
+	int n;
 
 	while (*p && isspace((unsigned char) *p))
 	    p++;
@@ -597,19 +624,25 @@ static int parse_rc_file(const char *filename)
 	    s = get_home_directory();
 	    strncpy(tmp, s, sizeof(tmp));
 	    p++;
-	    strncat(tmp, p, sizeof(tmp));
+	    safe_strncat(tmp, p, sizeof(tmp));
 	    strncpy(name, tmp, sizeof(name));
 	}
 
-	if (open_module(name, &idx))
+	if (open_module(name, &mod))
 	    continue;
 
 	if ((my_argv = parseargv(options, __progname, &my_argc)) == NULL)
 	    continue;
 
 	optind = 0;
+	n = init_module_options(my_argc, my_argv, mod);
 
-	if (init_module_options(my_argc, my_argv, modules[idx])) {
+	for (ap = my_argv; *ap; ap++)
+	    free(*ap);
+
+	free(my_argv);
+
+	if (n) {
 	    fclose(fp);
 	    return 2;
 	}
@@ -623,7 +656,6 @@ static int parse_rc_file(const char *filename)
 
 int main(int argc, char *argv[])
 {
-    int i = 0;
     int ret = EXIT_SUCCESS;
     int opt;
     char line[LINE_MAX], *s = NULL;
@@ -638,6 +670,8 @@ int main(int argc, char *argv[])
     chain_output = 1;
 
     while ((opt = getopt(argc, argv, "+x:X:dm:c:hO:F:t:vVLf")) != -1) {
+	struct module_s *mod;
+
 	/*
 	 * See getopt(3). 
 	 */
@@ -645,10 +679,8 @@ int main(int argc, char *argv[])
 
 	switch (opt) {
 	    case 'd':
-		i = module_index;
-
-		if (open_module("passwd.so", &i) == 0) {
-		    if (init_module_options(1, NULL, modules[i]))
+		if (open_module("passwd.so", &mod) == 0) {
+		    if (init_module_options(1, NULL, mod))
 			want_help = 1;
 		}
 		else {
@@ -656,8 +688,8 @@ int main(int argc, char *argv[])
 		    goto cleanup;
 		}
 
-		if (open_module("mail.so", &i) == 0) {
-		    if (init_module_options(1, NULL, modules[i]))
+		if (open_module("mail.so", &mod) == 0) {
+		    if (init_module_options(1, NULL, mod))
 			want_help = 1;
 		}
 		else {
@@ -665,8 +697,8 @@ int main(int argc, char *argv[])
 		    goto cleanup;
 		}
 
-		if (open_module("login.so", &i) == 0) {
-		    if (init_module_options(1, NULL, modules[i]))
+		if (open_module("login.so", &mod) == 0) {
+		    if (init_module_options(1, NULL, mod))
 			want_help = 1;
 		}
 		else {
@@ -726,12 +758,12 @@ int main(int argc, char *argv[])
 	    case 'x':
 		chaining = 1;
 	    case 'O':
-		if (open_module(optarg, &i)) {
+		if (open_module(optarg, &mod)) {
 		    ret = EXIT_FAILURE;
 		    goto cleanup;
 		}
 
-		if (init_module_options(argc, argv, modules[i]))
+		if (init_module_options(argc, argv, mod))
 		    want_help = 1;
 
 		/*
@@ -752,25 +784,26 @@ int main(int argc, char *argv[])
     }
 
     /* The last module cannot be chained (syntax). */
-    if (!module_index || TEST_FLAG(modules[module_index - 1].flags, MODULE_CHAINED))
+    if (!module_total || TEST_FLAG(modules[module_total-1]->flags, MODULE_CHAINED))
 	want_help = 1;
 
     /* Cycle through the modules and output their help text. */
     if (want_help) {
 	usage_header();
+	struct module_s **mod;
 
-	for (i = 0; i < module_index; i++) {
+	for (mod = modules; mod && *mod; mod++) {
 	    module_help *m_help;
 
-	    if (TEST_FLAG(modules[i].flags, MODULE_DUP))
+	    if (TEST_FLAG((*mod)->flags, MODULE_DUP))
 		continue;
 
-	    if ((m_help = dlsym(modules[i].m, "ui_module_help")) == NULL) {
+	    if ((m_help = dlsym((*mod)->m, "ui_module_help")) == NULL) {
 		warnx("%s", dlerror());
 		continue;
 	    }
 
-	    fprintf(stderr, "%s\n", modules[i].name);
+	    printf("%s\n", (*mod)->name);
 	    (*m_help) ();
 	}
 
